local temp_stack = {name="", count=0}

local function protect_entity(entity)
    entity.destructible = false;
    entity.minable = false;
end

local craftable_recipes = {
    "basic-oil-processing",
    "advanced-oil-processing",
    "sulfuric-acid",
    "heavy-oil-cracking",
    "light-oil-cracking",
    "solid-fuel-from-light-oil",
    "solid-fuel-from-petroleum-gas",
    "solid-fuel-from-heavy-oil",
    "sulfur",
    "plastic-bar",
    "battery",
    "copper-plate",
    "iron-plate",
    "steel-plate",
    "copper-cable",
    "iron-gear-wheel",
    "electronic-circuit",
    "advanced-circuit",
    "processing-unit",
    "radar",
    "solar-panel",
    "accumulator",
    "satellite",
    "speed-module",
    "rocket-control-unit",
    "rocket-fuel",
    "low-density-structure",
}

local tracked_items = {
    "iron-plate",
    "copper-plate",
    "electronic-circuit",
    "advanced-circuit",
    "processing-unit",
    "plastic-bar",
    "steel-plate",
    "speed-module",
    "rocket-control-unit",
    "rocket-fuel",
    "low-density-structure"
}

local tracked_liquids = {
    "petroleum-gas",
    "sulfuric-acid"
}

local tracked_needed = {
    ["iron-plate"] = 76587,
    ["copper-plate"] = 63588,
    ["electronic-circuit"] = 32005,
    ["advanced-circuit"] = 5240,
    ["processing-unit"] = 820,
    ["plastic-bar"] = 14580,
    ["steel-plate"] = 8700,
    ["speed-module"] = 720,
    ["rocket-control-unit"] = 720,
    ["rocket-fuel"] = 770,
    ["low-density-structure"] = 820,
    ["petroleum-gas"] = 23985,
    ["sulfuric-acid"] = 1410
}

local iron_belt, copper_belt, coal_belt, oil_pipe, forbidden_items, had_rocket_parts, repair_combinators;
local button_clicks = {};

local function create_planning_ui(player)
    local container = player.gui.top;
    if container.scenario_ui ~= nil then
        container.scenario_ui.destroy();
    end
    if container.production_ui ~= nil then
        container.production_ui.destroy();
    end
    local gui = container.add{type="flow", direction="horizontal", name="scenario_ui" }
    gui.add{type="button",caption="Run factory",name="run"}
end

local function update_production_ui(seconds)
    local tables;
    for _,player in pairs(game.connected_players) do
        if player.gui.top.production_ui then
            if tables == nil then tables = {} end;
            tables[player.index] = player.gui.top.production_ui.table;
        end
    end
    if tables == nil then return end;

    local force = game.forces.player;
    local update_labels = {};

    local function update_production_stat(name, amount)
        update_labels["produced-"..name] = amount;
        local ppm = 60*amount/seconds;
        if ppm >= 100 then ppm = math.floor(ppm)
        else ppm = math.floor(10*ppm)/10 end;
        update_labels["ppm-"..name] = ppm.."/m";
        update_labels["percent-"..name] = (math.floor(1000*amount/tracked_needed[name])/10).."%";
    end

    local item_stat = force.item_production_statistics;
    for _,name in ipairs(tracked_items) do
        update_production_stat(name, item_stat.get_input_count(name));
    end

    local fluid_stat = force.fluid_production_statistics;
    for _,name in ipairs(tracked_liquids) do
        update_production_stat(name, fluid_stat.get_input_count(name));
    end

    for _,table in pairs(tables) do
        for name,value in pairs(update_labels) do
            table[name].caption = value;
        end
    end
end

local function get_time_passed()
    local seconds = math.floor((game.tick - global.run_start_time)/60);
    local minutes = math.floor(seconds/60);
    local subseconds = seconds - minutes*60;
    local secstr = tostring(subseconds);
    if subseconds < 10 then
        secstr = "0"..secstr
    end;
    local minstr = tostring(minutes);
    if minutes < 10 then
        minstr = "0"..minstr
    end;
    return minstr .. ":" .. secstr, seconds;
end

local function create_production_ui(player)
    local container = player.gui.top;
    local frame = container.add{type="frame", direction="vertical", name="production_ui" }
    local count = #tracked_items + #tracked_liquids;
    local table = frame.add{type="table", name="table", colspan=count+1};
    local headers = {"Produced","Consumed","Needed"}
    table.add{type="label",caption="Item", name="base_header", style="frame_caption_label_style"}
    for id,name in ipairs(tracked_items) do
        table.add{type="sprite",name="h:"..name, sprite="item/"..name}
    end
    for id,name in ipairs(tracked_liquids) do
        table.add{type="sprite",name="h:"..name, sprite="fluid/"..name}
    end
    local function create_row(caption, prefix)
        table.add{type="label",caption=caption, name=prefix.."header", style="frame_caption_label_style" }
        for id,name in ipairs(tracked_items) do
            table.add{type="label", name=prefix..name}
        end
        for id,name in ipairs(tracked_liquids) do
            table.add{type="label", name=prefix..name}
        end
    end
    create_row("Produced","produced-");
    create_row("Per minute","ppm-");
    create_row("Needed *","needed-");
    create_row("Percentage","percent-");
    for name,count in pairs(tracked_needed) do
        table["needed-"..name].caption = count;
    end
    frame.add{type="label", name="disclaimer", caption="* Minimum amount of items required assuming 4x productivity 3 modules in the silo"};
    local _,seconds = get_time_passed();
    update_production_ui(seconds);
end

local function create_validating_ui(player)
    local container = player.gui.top;
    if container.scenario_ui ~= nil then
        container.scenario_ui.destroy();
    end
    local gui = container.add{type="frame", direction="vertical", name="scenario_ui" }
    gui.add{type="button",caption="Return to drawing board", name="reset"}
    gui.add{type="label",caption="Time", name="time", style="frame_caption_label_style"}
    local bar = gui.add{type="flow", direction="horizontal", name="speed_bar" }
    bar.add{type="button",caption="x1", name="x1"}
    bar.add{type="button",caption="x4", name="x4"}
    bar.add{type="button",caption="x15", name="x15"}
    bar.add{type="button",caption="MAX", name="turbo" }
    gui.add{type="label",caption="Status", name="status"}
    gui.add{type="button",caption="View production statistics", name="production"}
end

local function update_status()
    local time, seconds = get_time_passed();
    local status = "";

    if global.rocket_launched then
        status = "Launch!";
    elseif global.rocket_silo ~= nil and global.rocket_silo.valid then
        local inventory = global.rocket_silo.get_inventory(defines.inventory.rocket_silo_rocket);
        if inventory ~= nil then
            local item = inventory[1];
            if item.valid_for_read then
                if item.name == "satellite" then
                    status = "Launching!";
                    game.speed = 1;
                    global.rocket_launched = true;
                    global.rocket_silo.launch_rocket();
                else
                    status = "Wrong item in the satellite slot!"
                end
            else
                status = "Empty satellite slot"
            end
        elseif global.rocket_silo.recipe == nil then
            status = "test";
        else
            local parts = global.rocket_silo.rocket_parts;
            status = "Rocket parts: "..parts;
            if parts > 0 then had_rocket_parts = true end;
            if parts == 0 and had_rocket_parts then
                status = "Preparing...";
            end
        end
    else
        status = "No rocket silo";
    end

    for k,player in pairs(game.connected_players) do
        player.gui.top.scenario_ui.time.caption = "Time: "..time;
        player.gui.top.scenario_ui.status.caption = status;
    end

    if seconds % (game.speed * 2) == 0 then
        update_production_ui(seconds);
    end
end

local function set_player_character(player)
    player.set_controller{type=defines.controllers.character, character=global.characters[player.index] }
end

local function stop_simulating()
    if not global.running then return end;
    global.running = false;
    global.finished = nil;
    game.speed = 1;
    for k,player in pairs(game.connected_players) do
        set_player_character(player);
        create_planning_ui(player);
    end
end

local function clear_inventory(inventory)
    if inventory ~= nil then
        inventory.clear();
    end
end

local function recreate_entity(entity)
    local name = entity.name;
    local modules = entity.get_inventory(defines.inventory.assembling_machine_modules);
    local moduleDict;
    if modules ~= nil then
        moduleDict = modules.get_contents();
    end;
    local position = entity.position;
    local surface = entity.surface;
    entity.destroy();
    entity = surface.create_entity{name=name,position=position,force=game.forces.player }
    if moduleDict ~= nil then
        modules = entity.get_inventory(defines.inventory.assembling_machine_modules);
        for name,count in pairs(moduleDict) do
            modules.insert{name=name,count=count}
        end
    end
    return entity;
end

local dummy_signal = {type="item",name="coal"};
local combinator_clear_params = {
    ["arithmetic-combinator"] = {parameters={first_signal=dummy_signal,constant=0,operation="*",output_signal=dummy_signal}},
    ["decider-combinator"] = {parameters={first_signal=dummy_signal,constant=2147483647,operation=">",output_signal=dummy_signal}}
}

local function clear_map()
    -- Clear map: Reset all entities to default state and remove all forbidden items
    local surface = game.surfaces.nauvis;
    local coal_stack_insert = {name="coal",count=5};
    local coal_stack_insert_2 = {name="coal",count=1};
    repair_combinators = {};
    global.rocket_silo = nil;
    for k,entity in ipairs(surface.find_entities()) do
        if entity.valid then
            local type = entity.type;
            if type == "item-entity" then
                if forbidden_items[entity.stack.name] then
                    entity.destroy();
                end
            elseif type == "entity-ghost" then
                entity.destroy();
            elseif type == "arithmetic-combinator" or type == "decider-combinator" then
                -- to clear arithmetic combinator, temporarily set its operation to "* 0"
                -- to clear decider combinator, temporarily set its operation to "> int_max"
                local control = entity.get_control_behavior();
                if control then
                    repair_combinators[control] = control.parameters;
                    control.parameters = combinator_clear_params[type];
                end
            elseif type == "rocket-silo" then
                -- replace silo to fully reset its state
                entity = recreate_entity(entity);
                global.rocket_silo = entity;
            elseif entity.name == "steel-furnace" or type == "boiler" then
                -- There is no way to clear current fuel value, LuaEntity.energy does not work.
                -- So, we are recreating steel furnaces
                entity = recreate_entity(entity);
                entity.insert(coal_stack_insert)
            elseif type == "furnace" or type == "assembling-machine" then
                -- Clear specific inventories to preserve modules
                clear_inventory(entity.get_inventory(defines.inventory.assembling_machine_input))
                clear_inventory(entity.get_inventory(defines.inventory.assembling_machine_output))
                entity.crafting_progress = 0;
                entity.bonus_progress = 0;
            elseif type == "player" then
                -- nothing
            elseif type == "inserter" then
                if entity.held_stack.valid_for_read and forbidden_items[entity.held_stack.name] then
                    entity.held_stack.clear();
                end
                entity.energy = 0;
                if entity.name == "burner-inserter" then
                    clear_inventory(entity.get_inventory(defines.inventory.fuel))
                    entity.insert(coal_stack_insert_2)
                end
            else
                entity.energy = 0;
                if entity.has_items_inside() then
                    for item,_ in pairs(forbidden_items) do
                        local count = entity.get_item_count(item);
                        if count > 0 then
                            temp_stack.name = item;
                            temp_stack.count = count;
                            entity.remove_item(temp_stack);
                        end
                    end
                end
            end
            if entity.valid then
                for i=1,#entity.fluidbox do
                    entity.fluidbox[i] = nil;
                end
            end
        end
    end

    local item_stat = game.forces.player.item_production_statistics;
    for _,name in ipairs(tracked_items) do
        item_stat.set_input_count(name,0);
        item_stat.set_output_count(name,0);
    end

    local fluid_stat = game.forces.player.fluid_production_statistics;
    for _,name in ipairs(tracked_liquids) do
        fluid_stat.set_input_count(name,0);
        fluid_stat.set_output_count(name,0);
    end
end

local function start_simulating()
    if global.running then return end;
    global.running = true;
    global.rocket_launched = false;
    global.run_start_time = game.tick;
    had_rocket_parts = false;
    for k,player in pairs(game.connected_players) do
        create_validating_ui(player);
        player.set_controller{type=defines.controllers.ghost }
    end
    clear_map();
    update_status();
end

function button_clicks.run()
    start_simulating();
end

function button_clicks.reset()
    stop_simulating();
end

function button_clicks.x1()
    game.speed = 1;
end

function button_clicks.x4()
    game.speed = 4;
end

function button_clicks.x15()
    game.speed = 15;
end

function button_clicks.turbo()
    game.speed = 50;
end

function button_clicks.close_win_panel(index)
    local player = game.players[index];
    player.gui.center.win_panel.destroy();
end

function button_clicks.production(index)
    local player = game.players[index];
    if player.gui.top.production_ui then
        player.gui.top.production_ui.destroy();
    else
        create_production_ui(player);
    end
end

local function on_load()
    iron_belt = global.iron_belt;
    copper_belt = global.copper_belt;
    coal_belt = global.coal_belt;
    oil_pipe = global.oil_pipe;
    forbidden_items = global.forbidden_items;
end
script.on_load(on_load);

script.on_init(function()
    local force = game.forces.player;
    force.research_all_technologies();
    force.manual_crafting_speed_modifier = -1;
    force.character_logistic_slot_count = 0;
    force.character_trash_slot_count = 0;
    force.disable_research();

    game.disable_tips_and_tricks();
    local surface = game.surfaces.nauvis;
    global.iron_belt = surface.find_entity("fast-underground-belt",{-33.5,-21.5});
    global.copper_belt = surface.find_entity("fast-underground-belt",{-33.5,-9.5});
    global.coal_belt = surface.find_entity("underground-belt",{-33.5,2.5});
    global.oil_pipe = surface.find_entity("pipe-to-ground",{-33.5,14.5});
    global.forbidden_items = {["iron-ore"]=true,["copper-ore"]=true,["coal"]=true};
    global.characters = {};
    global.running = false;
    surface.always_day=true;
    on_load();
    protect_entity(iron_belt);
    protect_entity(copper_belt);
    protect_entity(coal_belt);
    protect_entity(oil_pipe);

    local recipes = force.recipes;
    for name,recipe in pairs(recipes) do
        recipe.enabled = false;
    end
    for k,name in ipairs(craftable_recipes) do
        local recipe = recipes[name];
        recipes[name].enabled = true;
        for k,product in pairs(recipe.products) do
            if product.type == "item" then
                forbidden_items[product.name] = true;
            end
        end
    end
end)

local function add_item_to_belt(belt, item)
    for i=1,2 do
        local line = belt.get_transport_line(i);
        if line.can_insert_at_back() then
            line.insert_at_back(item);
        end
    end
end

local function setup_player_base(player)
    local character = player.character;

    local armor = character.get_inventory(defines.inventory.player_armor)[1];
    armor.set_stack({name = "power-armor-mk2", count=1})
    local grid = armor.grid;
    grid.put{name="fusion-reactor-equipment", position = {0,0}}
    grid.put{name="fusion-reactor-equipment", position = {4,0}}
    grid.put{name="fusion-reactor-equipment", position = {0,4}}
    grid.put{name="fusion-reactor-equipment", position = {4,4}}
    grid.put{name="exoskeleton-equipment", position = {8,0}}
    grid.put{name="exoskeleton-equipment", position = {8,4}}
    grid.put{name="personal-roboport-equipment", position = {0,8}}
    grid.put{name="personal-roboport-equipment", position = {2,8}}
    grid.put{name="personal-roboport-equipment", position = {4,8}}
    grid.put{name="personal-roboport-equipment", position = {6,8}}
    grid.put{name="personal-roboport-equipment", position = {8,8}}

    local bluebook = character.get_inventory(defines.inventory.player_quickbar)[1];
    bluebook.set_stack{name = "blueprint-book", count=1}
    local blueprints = bluebook.get_inventory(defines.inventory.item_main);
    for i=1,10 do
        blueprints.insert{name = "blueprint", count = 1};
    end

    character.insert{name="steel-axe", count=20};
    character.insert{name="construction-robot", count=50};
    character.insert{name="deconstruction-planner", count=1};
end

local function setup_player_inventory(player)
    local character = player.character

    character.insert{name="transport-belt", count=800}
    character.insert{name="fast-transport-belt", count=150}
    character.insert{name="underground-belt", count=150}
    character.insert{name="fast-underground-belt", count=20}
    character.insert{name="splitter", count=100}
    character.insert{name="fast-splitter", count=10}
    character.insert{name="inserter", count=400}
    character.insert{name="long-handed-inserter", count=200}
    character.insert{name="fast-inserter", count=150}
    character.insert{name="filter-inserter", count=50}
    character.insert{name="medium-electric-pole", count=150}
    character.insert{name="pipe", count=300}
    character.insert{name="pipe-to-ground", count=150}
    character.insert{name="assembling-machine-2", count=100}
    character.insert{name="steel-furnace", count=150 }
    character.insert{name="boiler", count=100}
    character.insert{name="steam-engine", count=50 }
    character.insert{name="offshore-pump", count=10}
    character.insert{name="assembling-machine-3", count=20}
    character.insert{name="steel-chest", count=50}
    character.insert{name="substation", count=10}
    character.insert{name="electric-furnace", count=50}
    character.insert{name="small-pump", count=20}
    character.insert{name="red-wire", count=200}
    character.insert{name="green-wire", count=200}
    character.insert{name="arithmetic-combinator", count=20}
    character.insert{name="decider-combinator", count=20}
    character.insert{name="constant-combinator", count=20}
    character.insert{name="oil-refinery", count=10}
    character.insert{name="chemical-plant", count=30}
    character.insert{name="storage-tank", count=10}
    character.insert{name="effectivity-module", count=50}
    character.insert{name="speed-module-2", count=20}
    character.insert{name="productivity-module-3", count=4}
    character.insert{name="rocket-silo", count=1}
    character.insert{name="burner-inserter", count=50 }
    character.insert{name="stack-inserter", count=5}
end

local iron_stack = {name="iron-ore", count=1}
local copper_stack = {name="copper-ore", count=1}
local coal_stack = {name="coal", count=1 }

--[[
local iron_stack = {name="rocket-control-unit", count=1 }
local copper_stack = {name="low-density-structure", count=1 }
local coal_stack = {name="rocket-fuel", count=1}
]]

local function remove_forbidden_items(event)
    local player = game.players[event.player_index];
    for name,_ in pairs(forbidden_items) do
        local count = player.get_item_count(name);
        if name == "coal" then count = count - 50 end; -- Allow player to hold 50 coal
        if count > 0 then
            temp_stack.name = name;
            temp_stack.count = count;
            player.remove_item(temp_stack);
        end
    end
end

script.on_event(defines.events.on_player_main_inventory_changed, remove_forbidden_items);
script.on_event(defines.events.on_player_quickbar_inventory_changed, remove_forbidden_items);

script.on_event(defines.events.on_tick, function(event)
    local tick = event.tick;
    if tick % 5 == 0 then
        add_item_to_belt(iron_belt, iron_stack);
        add_item_to_belt(copper_belt, copper_stack);
    end
    if tick % 10 == 0 then
        add_item_to_belt(coal_belt, coal_stack);
    end
    local oil = oil_pipe.fluidbox[1];
    if oil == nil or oil.type ~= "crude-oil" then
        oil = {type="crude-oil", amount = 0.16666667}
    else
        oil.amount = oil.amount + 0.16666667;
    end
    oil_pipe.fluidbox[1] = oil;

    if global.running and global.finished == nil then
        if global.run_start_time % 60 == game.tick % 60 then
            update_status();
        end;
        if repair_combinators and game.tick > global.run_start_time then
            for control, value in pairs(repair_combinators) do
                if control.valid then control.parameters = value end;
            end
            repair_combinators = nil;
        end
    end
end)

script.on_event(defines.events.on_player_created, function(event)
    local player = game.players[event.player_index]
    player.minimap_enabled = false;
    player.disable_recipe_groups();
    player.game_view_settings.show_research_info = false;
    setup_player_base(player)
    if global.player_setup == nil then
        setup_player_inventory(player)
        global.player_setup = true
    end;
end)

script.on_event(defines.events.on_player_joined_game, function(event)
    local player = game.players[event.player_index]
    if player.character then
        global.characters[player.index] = player.character;
    end
    if global.running then
        create_validating_ui(player);
        player.set_controller{type=defines.controllers.ghost }
    else
        set_player_character(player);
        create_planning_ui(player);
    end
end);

script.on_event(defines.events.on_gui_click, function(event)
    local element = event.element;
    local handler = button_clicks[element.name];
    if handler ~= nil then
        handler(event.player_index);
    end
end)

script.on_event(defines.events.on_rocket_launched, function(event)
    if not global.running or not global.rocket_launched then return end;
    global.finished = true;
    if not global.won then
        game.set_game_state{game_finished = true, player_won = true, can_continue = true }
        global.won = true;
    end
    for _,player in pairs(game.connected_players) do
        local container = player.gui.center;
        local frame = container.add{type="frame", direction="vertical", name="win_panel" }
        frame.add{type="label",caption="Congratulations!", name="header", style="frame_caption_label_style" }
        frame.add{type="label",caption="Your time is "..get_time_passed(), name="time", style="frame_caption_label_style" }
        frame.add{type="button",caption="Close", name="close_win_panel"}
    end
end)